package game.mvc.controllers 
{
    import flash.display.DisplayObjectContainer;
    import flash.display.Stage;
    import flash.events.Event;
    import flash.events.MouseEvent;
    import flash.events.TimerEvent;
    import flash.utils.Timer;
    import game.constants.Size;
    import game.mvc.model.BaseModel;
    import game.mvc.model.bullets.BulletModel;
    import game.mvc.model.circle.CirclesModel;
    import game.mvc.model.circle.CircleModel;
    import game.mvc.model.enemy.EnemyModel;
    import game.mvc.model.tank.TankModel;
    import game.mvc.views.game.GameView;
    import game.mvc.views.score.ScoreView;
    import game.mvc.views.windows.RestartWindowView;
    import game.utils.GeometricUtils;
    
    public class BaseController 
    {
        private static const ENEMY_RESPAWN_DELAY:int = 200;
        private static const SHOOT_TIMER:int = 100;
        
        private var container:DisplayObjectContainer;
        private var baseModel:BaseModel;
        
        private var game:GameView;
        private var score:ScoreView;
        private var restartWindow:RestartWindowView;
        
        private var keysController:KeysController;
        private var respawnTimer:Timer = new Timer(ENEMY_RESPAWN_DELAY);
        private var shootTimer:Timer = new Timer(SHOOT_TIMER, 1);
        
        private var canShoot:Boolean;
        private var mouseDone:Boolean;
        
        public function BaseController(container:DisplayObjectContainer)
        {
            this.container = container;
            init();
        }
        
        private function init():void 
        {
            createGameObjects();
            start();
        }
        
        private function createGameObjects():void 
        {
            baseModel = new BaseModel();
            game = new GameView(baseModel);
            container.addChild(game);
            score = new ScoreView(baseModel.score);
            container.addChild(score);
            restartWindow = new RestartWindowView();
            keysController = new KeysController(container.stage);
        }
        
        private function start():void
        {
            canShoot = true;
            mouseDone = false;
            respawnTimer.reset();
            shootTimer.reset();
            respawnTimer.addEventListener(TimerEvent.TIMER, respawnTimerListener);
            shootTimer.addEventListener(TimerEvent.TIMER_COMPLETE, shootTimerListener)
            game.addEventListener(MouseEvent.MOUSE_DOWN, mouseDownListener);
            game.addEventListener(MouseEvent.MOUSE_UP, mouseUp);
            container.addEventListener(Event.ENTER_FRAME, enterFrameListener);
            respawnTimer.start();
            game.update();
        }
        
        private function respawnTimerListener(e:TimerEvent):void 
        {
            var enemy:EnemyModel = new EnemyModel();
            baseModel.enemies.add(enemy);
            game.addEnemy(enemy);
        }
        
        private function shootTimerListener(e:TimerEvent):void 
        {
            canShoot = true;
            tryShoot();
        }
        
        private function mouseDownListener(e:MouseEvent):void 
        {
            mouseDone = true;
            tryShoot();
        }
        
        private function mouseUp(e:MouseEvent):void 
        {
            mouseDone = false;
        }
        
        private function tryShoot():void
        {
            if (mouseDone && canShoot)
                shoot();
        }
        
        private function shoot():void
        {
            var tank:TankModel = baseModel.tank;
            var bullet:BulletModel = new BulletModel(tank.x, tank.y, tank.rotation);
            baseModel.bullets.add(bullet);
            game.addBullet(bullet);
            canShoot = false;
            shootTimer.reset();
            shootTimer.start();
        }
        
        private function enterFrameListener(e:Event):void 
        {
            updateTank();
            updateCircles(baseModel.enemies);
            updateCircles(baseModel.bullets);
            testBulletsIntersection();
            testTankAndEmenyInterection();
            game.update();
        }
        
        private function updateTank():void
        {
            keysController.update();
            var tank:TankModel = baseModel.tank;
            if (keysController.isLeft) tank.moveLeft();
            if (keysController.isRight) tank.moveRight();
            if (keysController.isDown) tank.moveDown();
            if (keysController.isUp) tank.moveUp();
            if (keysController.isUpLeft) tank.moveUpLeft();
            if (keysController.isUpRight) tank.moveUpRight();
            if (keysController.isDownLeft) tank.moveDownLeft();
            if (keysController.isDownRight) tank.moveDownRight();
            var stage:Stage = container.stage;
            tank.updateRotation(stage.mouseX, stage.mouseY);
        }
        
        private function updateCircles(circles:CirclesModel):void 
        {
            for (var i:int = 0; i < circles.length; i++) 
            {
                var circle:CircleModel = circles.getCircle(i);
                circle.move();
                if (circle.isEnd)
                {
                    circles.deleteByIndex(i);
                    i--;
                }
            }
        }
        
        private function testBulletsIntersection():void 
        {
            var bullets:CirclesModel = baseModel.bullets;
            for (var i:int = 0; i < bullets.length; i++) 
            {
                var bullet:CircleModel = bullets.getCircle(i);
                var kill:Boolean = killEnemyFromBullet(bullet);
                if (kill)
                {
                    bullets.deleteByIndex(i);
                    i--;
                }
            }
        }
        
        private function killEnemyFromBullet(bullet:CircleModel):Boolean
        {
            var enemies:CirclesModel = baseModel.enemies;
            for (var i:int = 0; i < enemies.length; i++) 
            {
                var enemy:EnemyModel = enemies.getCircle(i) as EnemyModel;
                var distance:Number = GeometricUtils.getDistance(bullet.x, bullet.y, enemy.x, enemy.y);
                if (distance < Size.ENEMY_RADIUS + Size.BULLET_RADIUS)
                {
                    baseModel.score.add(enemy.score);
                    enemies.deleteByIndex(i);
                    return true;
                }
            }
            return false;
        }
        
        private function testTankAndEmenyInterection():void 
        {
            var enemies:CirclesModel = baseModel.enemies;
            var tank:TankModel = baseModel.tank;
            for (var i:int = 0; i < enemies.length; i++) 
            {
                var enemy:CircleModel = enemies.getCircle(i);
                var distance:Number = GeometricUtils.getDistance(tank.x, tank.y, enemy.x, enemy.y);
                if (distance < Size.ENEMY_RADIUS + Size.PLAYER_RADIUS)
                    stopAndShowRestartWindow();
            }
        }
        
        private function stopAndShowRestartWindow():void
        {
            respawnTimer.stop();
            shootTimer.stop();
            respawnTimer.removeEventListener(TimerEvent.TIMER, respawnTimerListener);
            shootTimer.removeEventListener(TimerEvent.TIMER_COMPLETE, shootTimerListener)
            container.removeEventListener(Event.ENTER_FRAME, enterFrameListener);
            game.removeEventListener(MouseEvent.MOUSE_DOWN, mouseDownListener);
            game.removeEventListener(MouseEvent.MOUSE_UP, mouseUp);
            container.addChild(restartWindow);
            restartWindow.addEventListener(MouseEvent.CLICK, restartWindowClickListener);
        }
        
        private function restartWindowClickListener(e:MouseEvent):void 
        {
            restartWindow.removeEventListener(MouseEvent.CLICK, restartWindowClickListener);
            container.removeChild(restartWindow);
            baseModel.restart();
            game.restart();
            start();
        }
    }
}